# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qrdm.proto
# Protobuf Python Version: 4.25.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\nqrdm.proto\x12\x04qrdm"\xaa\x01\n\tQRContent\x12$\n\x04meta\x18\x01 \x01(\x0b\x32\x16.qrdm.QRContent.QRMeta\x12\x14\n\x0c\x64oc_fragment\x18\x02 \x01(\x0c\x1a\x61\n\x06QRMeta\x12\x15\n\rdocument_hash\x18\x01 \x01(\x06\x12\x17\n\x0fsequence_number\x18\x02 \x01(\r\x12\x16\n\x0etotal_qr_codes\x18\x03 \x01(\r\x12\x0f\n\x07num_ecc\x18\x04 \x01(\r"\x84\x01\n\x0f\x44ocumentPayload\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\x12\x10\n\x08metadata\x18\x02 \x01(\t\x12\x31\n\tdata_type\x18\x03 \x01(\x0e\x32\x1e.qrdm.DocumentPayload.DataType"\x1b\n\x08\x44\x61taType\x12\x0f\n\x0bUTF8_STRING\x10\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "qrdm_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS is False:
    DESCRIPTOR._options = None
    _globals["_QRCONTENT"]._serialized_start = 21
    _globals["_QRCONTENT"]._serialized_end = 191
    _globals["_QRCONTENT_QRMETA"]._serialized_start = 94
    _globals["_QRCONTENT_QRMETA"]._serialized_end = 191
    _globals["_DOCUMENTPAYLOAD"]._serialized_start = 194
    _globals["_DOCUMENTPAYLOAD"]._serialized_end = 326
    _globals["_DOCUMENTPAYLOAD_DATATYPE"]._serialized_start = 299
    _globals["_DOCUMENTPAYLOAD_DATATYPE"]._serialized_end = 326
# @@protoc_insertion_point(module_scope)
